/* This file is a part of cx2450x.c and will be included directly */

/*******************************************************************************/
/* fuction block for accessing the GXA registers                               */
/*******************************************************************************/

static u32 _read_gxa(volatile u8 *base_addr, u32 offset)
{
    u32 wait_cmp = 0x40000000;	/* default, register access */

    if (offset & 0x0003F800)	/* command access */
	wait_cmp = 0x4E010000;

    while(1)
    {
	u32 depth = (*(volatile u32 *)(base_addr + GXA_DEPTH_REG));
	if (depth & wait_cmp)
	continue;
	break;
    }

    return *(volatile u32 *)(base_addr + offset);
}

/*******************************************************************************/

static void _write_gxa(volatile u8 *base_addr, u32 offset, u32 value)
{
    u32 wait_cmp = 0x40000000;	/* default, register access */

    if (offset & 0x0003F800)	/* command access */
	wait_cmp = 0x4E040000;

    while(1) 
    {
	u32 depth = *((volatile u32 *)(base_addr + GXA_DEPTH_REG));
	if (depth & wait_cmp)
	    continue;
	break;
    }

    *(volatile u32 *)(base_addr + offset) = value;
}

/*******************************************************************************/

static DFBResult _cx2450x_sync_engine(void *driver_data, void *device_data)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    u32 depth = *((volatile u32 *)(drv->gxa_base + GXA_DEPTH_REG));

    if ((depth & 0x01000000) == 0)
	_write_gxa(drv->gxa_base, GXA_CMD_QMARK | GXA_PARAM_COUNT(1), 0x0100);

    return DFB_OK;
}

/*******************************************************************************/
/* all the accelleration functions                                             */
/*******************************************************************************/

bool _cx2450x_draw_line(void *driver_data, void *device_data, DFBRegion *line)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;

    u32 cmd = GXA_CMD_NOT_TEXT | GXA_SRC_BMP_SEL(2) | GXA_DST_BMP_SEL(2) | dev->cmd_opts;

    _write_gxa(drv->gxa_base, GXA_LINE_CONTROL_REG, 0x00000000);
    _write_gxa(drv->gxa_base, cmd | GXA_PARAM_COUNT(2), GXA_POINT(line->x1, line->y1));
    _write_gxa(drv->gxa_base, cmd | GXA_PARAM_COUNT(2), GXA_POINT(line->x2, line->y2));

    return true;
}

/*******************************************************************************/

static bool _cx2450x_draw_rectangle(void *driver_data, void *device_data, DFBRectangle *rect)
{
    /* the accelerator can handle multi lines. This means, with the first command, we setup
       the start and endpoint for the first line to draw. The GXA will automatically setup
       end point as the start point of the new line. So we only need to setup the new end point. */

    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;

#if 0
    /* Here we do not use the things written above. Drawing 4 lines with their start and end point
       is faster than just passing next endpoint. Seems the GXA lacks performance, if we draw
       backward (from right to left anf from bottom to top). */

    u32 cmd = GXA_CMD_NOT_TEXT | GXA_SRC_BMP_SEL(2) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(2) | dev->cmd_opts;
    u32 xr = rect->x + rect->w;
    u32 yl = rect->y + rect->h;

    /* draw the horizontal lines */
    _write_gxa(drv->gxa_base, GXA_LINE_CONTROL_REG, 0x00000404);	/* X is major, skip last pixel */
    
    /* upper left to upper right */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(xr, rect->y));		/* endig point */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y));	/* start point */

    /* lower left to lower right */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(xr, yl));			/* endig point */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, yl));		/* start point */

    /* draw the vertical lines */
    _write_gxa(drv->gxa_base, GXA_LINE_CONTROL_REG, 0x00000004);	/* Y is major, skip last pixel */

    /* upper left to lower left */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, yl));		/* endig point */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y));	/* start point */

    /* upper right to lower right */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(xr, yl));			/* endig point */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(xr, rect->y));		/* start point */
#endif

    /* This is a solution to draw rectangles by doing 4 solid fills with the blitter. Doing this
       increase the performance by around 8 percent (3 percent with blend) */

    u32 cmd = GXA_CMD_BLT | GXA_CMD_NOT_TEXT | GXA_SRC_BMP_SEL(7) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(2) | dev->cmd_opts;

    /* upper H-line */
    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y));		/* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->w, 1));			/* width/hight */

    /* lower H-line */
    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y + rect->h - 1));	/* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->w, 1));			/* width/hight */

    /* left V-line */
    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y));		/* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(1, rect->h));			/* width/hight */

    /* right V-line */
    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x + rect->w - 1, rect->y));	/* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(1, rect->h));			/* width/hight */

    return true;
}

/*******************************************************************************/

static bool _cx2450x_fill_rectangle(void *driver_data, void *device_data, DFBRectangle *rect)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;

    u32 cmd = GXA_CMD_BLT | GXA_CMD_NOT_TEXT | GXA_SRC_BMP_SEL(7) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(2) | dev->cmd_opts;

    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y));	/* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->w, rect->h));	/* width/hight */

    return true;
}

/*******************************************************************************/

//static bool _cx2450x_fill_triangle(void *driver_data, void *device_data, DFBTriangle *tri)
//{
    /* unsupported */
//    return false;
//}

/*******************************************************************************/

static bool _cx2450x_blit(void *driver_data, void *device_data, DFBRectangle *rect, int dx, int dy)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;
    u32 cmd = GXA_CMD_BLT | GXA_CMD_NOT_TEXT | GXA_SRC_BMP_SEL(1) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(3) | dev->cmd_opts;

    if (dev->b_soft)
	return false;

    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(dx, dy));             /* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->w, rect->h));   /* source width/hight */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(rect->x, rect->y));   /* source x/y */

    return true;
}

/*******************************************************************************/

static bool _cx2450x_stretchblit(void *driver_data, void *device_data, DFBRectangle *srect, DFBRectangle *drect)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;
    u32 cmd = GXA_CMD_STRETCH_BLT | GXA_CMD_BLT | GXA_CMD_NOT_TEXT | GXA_SRC_BMP_SEL(1) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(7) | dev->cmd_opts;
    u32 ys = 1, xs = 1, sp = 1;
    u32 hco, vco;

    if (dev->b_soft)
	return false;

    if (dev->cmd_opts & GXA_CMD_NOT_ALPHA)
        cmd |= GXA_CMD_ROP;

    /* prevent divide by 0 */
    if (drect->h)
        ys = ((srect->h << 16) / drect->h) & 0x00FFFFFF;
    if (drect->w)
        xs = ((srect->w << 16) / drect->w) & 0x00FFFFFF;
    if (srect->w)
        sp = (((drect->w << 20) / srect->w) * 3) & 0x0FFFFFFF;

    if (srect->w <= drect->w)
	hco = 0;
    else if (srect->w <= (drect->w << 1))
	hco = 0x0100;
    else if (srect->w <= (drect->w << 2))
	hco = 0x0200;
    else
	hco = 0x0300;

    if (srect->h <= drect->h)
	vco = 0;
    else if (srect->h <= (drect->h << 1))
	vco = 0x0100;
    else if (srect->h <= (drect->h << 2))
	vco = 0x0200;
    else
	vco = 0x0300;

    _write_gxa(drv->gxa_base, GXA_CMD_HFILTER | GXA_SRC_BMP_SEL(1) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(1), (u32)(drv->coeff_table + hco));
    _write_gxa(drv->gxa_base, GXA_CMD_VFILTER | GXA_SRC_BMP_SEL(1) | GXA_DST_BMP_SEL(2) | GXA_PARAM_COUNT(1), (u32)(drv->coeff_table + vco));

    _write_gxa(drv->gxa_base, GXA_BLT_CONTROL_REG, 0);

    _write_gxa(drv->gxa_base, cmd, GXA_POINT(drect->x, drect->y)); /* destination x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(drect->w, drect->h)); /* destination width/hight */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(srect->x, srect->y)); /* source x/y */
    _write_gxa(drv->gxa_base, cmd, GXA_POINT(srect->w, srect->h)); /* source width/hight */
    _write_gxa(drv->gxa_base, cmd, ys);                            /* Y sclae factor */
    _write_gxa(drv->gxa_base, cmd, xs);                            /* X sclae factor */
    _write_gxa(drv->gxa_base, cmd, sp);                            /* destination X patch */

    return true;
}

/*******************************************************************************/
