/* This file is a part of cx2450x.c and will be included directly */

static u32 _make_cx2450x_bmp_reg(DFBSurfacePixelFormat dfbcolorformat, u32 pitch)
{
    u32 reg = 0;

    switch (DFB_BITS_PER_PIXEL(dfbcolorformat))
    {
        case 32:                                /* 32 bpp */
            reg = (3 << 16) | (pitch / 4);
            break;
        case 16:                                /* 16 bpp */
            reg = (2 << 16) | (pitch / 2);
            break;
        case 8:                                 /* 8 bpp */
            reg = (1 << 16) |  pitch;
            break;
        default:
            D_INFO("%s() unhandled %d bpp format\n", __FUNCTION__, DFB_BITS_PER_PIXEL(dfbcolorformat));
            reg = (1 << 16) |  pitch;
            break;
    }

    switch (dfbcolorformat)
    {
	case DSPF_RGB32:
	case DSPF_ARGB:
	    reg |= 0x03 << 16;	/* 32 bit ARGB (4 byte, alpha 8@24, R 8@16, G 8@8, B 8@0) */
	    break;
	case DSPF_RGB16:        /* 16 bit RGB  (2 byte, red 5@11, green 6@5, blue 5@0) */
	    reg |= 0x08 << 16;
	    break;
	case DSPF_ARGB1555:     /* 16 bit ARGB (2 byte, alpha 1@15, red 5@10, green 5@5, blue 5@0) */
	    reg |= 0x10 << 16;
	    break;
        case DSPF_AYUV:         /* 32 bit AYUV (4 byte, alpha 8@24, Y 8@16, Cb 8@8, Cr 8@0) */
	    reg |= 0x04 << 16;
	    reg |= 3 << 21;	/* Y and C in range from 0 to 255 */
	    break;
	default:
            break;
    }
																																		
    return reg;
}

/*******************************************************************************/

static u32 _make_cx2450x_color(DFBColor color, DFBSurfacePixelFormat format)
{
    u32 cx_color = 0;

    switch (format)
    {
	case DSPF_RGB32:	/* 32 bit RGB without alpha channel */
	    /* do the same as for ARGB - DFB passes alpha values for blending, so do not drop/overwrite alpha */
	case DSPF_ARGB:		/* 32 bit RGB including alpha channel */
	    cx_color = (color.a << 24) | (color.r << 16) | (color.g << 8) | color.b;
	    break;
	case DSPF_AYUV:
	    cx_color = (color.a << 24) | (color.r << 16) | (color.b << 8) | color.g;	/* not sure if that conversion is currect */
	    break;
	default:
            D_INFO("%s() unhandled color format %.8X\n", __FUNCTION__, format);
	    break;
    }

    return cx_color;
}

/*******************************************************************************/

static void _cx2450x_check_state(void *driver_data, void *device_data, CardState *state, DFBAccelerationMask accel)
{
//D_INFO("%s()\n", __FUNCTION__);

    switch (state->destination->config.format)
    {
        /* supported color formats - the HW knows more, but they does not fit to DFB's formats (espeically YUV) */
	case DSPF_RGB32:	/* 24 bit   RGB (4 byte, nothing@24, red 8@16, green 8@8, blue 8@0) */
        case DSPF_ARGB:		/* 32 bit  ARGB (4 byte, alpha 8@24, red 8@16, green 8@8, blue 8@0) */
            break;
        default:
    	    /* all the DFB formats, which are not handled by the HW directly  */
//	    D_INFO("%s() unhandled destination format 0x%.8X\n", __FUNCTION__, state->destination->config.format);
            return;
    }

    if (DFB_DRAWING_FUNCTION(accel))
    {
        if (state->drawingflags & ~CX2450X_SUPPORTED_DRAWING_FLAGS)
	{
//	    D_INFO("%s() [DRAW] unhandled drawing flag(s) 0x%.8X [0x%.8X]\n", __FUNCTION__, state->drawingflags & ~CX2450X_SUPPORTED_DRAWING_FLAGS, CX2450X_SUPPORTED_DRAWING_FLAGS);
            return;
	}

        state->accel |= CX2450X_SUPPORTED_DRAWING_FUNCTIONS;
    }

    if (DFB_BLITTING_FUNCTION(accel))
    {
        if (state->blittingflags & ~CX2450X_SUPPORTED_BLITTING_FLAGS)
	{
//	    D_INFO("%s() [BLIT] unhandled blitting flag(s) 0x%.8X [0x%.8X]\n", __FUNCTION__, state->blittingflags & ~CX2450X_SUPPORTED_BLITTING_FLAGS, CX2450X_SUPPORTED_BLITTING_FLAGS);
            return;
	}
	
        switch (state->source->config.format)
        {
	    case DSPF_A8:	/* Rendered text comes in this alpha only format */
	    case DSPF_LUT8:	/* Rendered text with blend comes in LUT8 */
	    case DSPF_RGB16:
            case DSPF_RGB32:    /* modify alpha to a solid value (0xFF) */
            case DSPF_ARGB:
                break;
            default:
		D_INFO("%s() [BLIT] unhandled source format 0x%.8X\n", __FUNCTION__, state->source->config.format);
                return;
        }
        state->accel |= CX2450X_SUPPORTED_BLITTING_FUNCTIONS;
    }
}

/*******************************************************************************/

static void _cx2450x_set_state(void *driver_data, void *device_data, GraphicsDeviceFuncs *funcs, CardState *state, DFBAccelerationMask accel)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;

    if (state->mod_hw)
    {
//	D_INFO("%s() state->mod_hw: 0x%.8X, accel: 0x%.8X\n", __FUNCTION__, state->mod_hw, accel);

        switch (accel)
	{
    	    case DFXL_FILLRECTANGLE:			/* 0x00000001 */
	    case DFXL_DRAWRECTANGLE:			/* 0x00000002 */
	    case DFXL_DRAWLINE:				/* 0x00000004 */
	    {
		u32 cfg_reg = DEF_CFG_REG;

        	if (state->mod_hw & SMF_DRAWING_FLAGS)			/* 0x00000001 */
		{
		    dev->cmd_opts = GXA_CMD_NOT_ALPHA;

            	    if (state->drawingflags & DSDRAW_BLEND)		/* 0x00000001 - uses alpha from color */
		    {
			dev->cmd_opts &= ~GXA_CMD_NOT_ALPHA;
			_write_gxa(drv->gxa_base, GXA_BLEND_CFG_REG, (0x0C << 16) | (0x0D << 12) | (0x0C << 6) | (0x0D << 2));
			/* no need to setup color here. It's done below with SMF_COLOR */
		    }

		    if (state->drawingflags & DSDRAW_DST_COLORKEY)	/* 0x00000002 - write to destination only if the destination pixel matches the destination color key */
		    {
		    
		    }

		    if (state->drawingflags & DSDRAW_SRC_PREMULTIPLY)	/* 0x00000004 - multiplies the color's rgb channels by the alpha channel before drawing */
		    {
		    
		    }

		    if (state->drawingflags & DSDRAW_DST_PREMULTIPLY)	/* 0x00000008 - modulates the dest. color with the dest. alpha */
		    {
		    
		    }

		    if (state->drawingflags & DSDRAW_DEMULTIPLY)	/* 0x00000010 - divides the color by the alpha before writing the data to the destination */
		    {
		    
		    }

		    if (state->drawingflags & DSDRAW_XOR)		/* 0x00000020 - bitwise xor the destination pixels with the specified color after premultiplication */
		    {
		    
		    }
		}

        	if (state->mod_hw & SMF_COLOR)				/* 0x00000008 */
	        {
#warning check
		    if ((state->drawingflags == 0) && (state->destination->config.format == DSPF_RGB32))			/* force alpha to 0xFF */
			state->color.a = 0xFF;

		    _write_gxa(drv->gxa_base, GXA_FG_COLOR_REG, _make_cx2450x_color(state->color, state->destination->config.format));
		    _write_gxa(drv->gxa_base, GXA_BMP7_TYPE_REG, _make_cx2450x_bmp_reg(state->destination->config.format, state->dst.pitch) | (1 << 27));
		    _write_gxa(drv->gxa_base, GXA_BG_COLOR_REG, _make_cx2450x_color(state->color, state->destination->config.format));
		    _write_gxa(drv->gxa_base, GXA_CONST_COLOR_REG, _make_cx2450x_color(state->color, state->destination->config.format));
		    _write_gxa(drv->gxa_base, GXA_CFG2_REG, DEF_CFG2_REG | (state->color.a & 0xFF));								/* blend */
		}

		if (state->mod_hw & SMF_DESTINATION)			/* 0x00000100 */
    		{
		    /* setup the destination bitmap - as supposed by the datasheet, we use bitmap context #2 */
	    	    _write_gxa(drv->gxa_base, GXA_BMP2_TYPE_REG, _make_cx2450x_bmp_reg(state->destination->config.format, state->dst.pitch));
		    _write_gxa(drv->gxa_base, GXA_BMP2_ADDR_REG, drv->smem_start + state->dst.offset);
		}

		_write_gxa(drv->gxa_base, GXA_CFG_REG, cfg_reg);
		break;
	    }
	    case DFXL_BLIT:
	    case DFXL_STRETCHBLIT:
	    {
		if (state->mod_hw & SMF_BLITTING_FLAGS)	/* 0x00000002 */
        	{
		    u32 cfg_reg   = DEF_CFG_REG;
		    u32 blend_reg = 0;

		    dev->b_flags    = state->blittingflags;
        	    dev->b_soft     = false;
		    dev->cmd_opts = GXA_CMD_NOT_ALPHA;

                    if (state->blittingflags & DSBLIT_BLEND_ALPHACHANNEL)	/* 0x00000001 - enables blending and uses alphachannel from source */
		    {
			dev->cmd_opts &= ~GXA_CMD_NOT_ALPHA;
			_write_gxa(drv->gxa_base, GXA_CFG2_REG, DEF_CFG2_REG & 0xFFFFFF00);
			if (state->destination->config.format == DSPF_RGB32)
			    dev->b_soft = true;
			else
			    blend_reg = (0x08 << 16) | (0x09 << 12) | (0x08 << 6) | (0x09 << 2);	// ((0x08 << 16) | (0x09 << 12) | (0x08 << 6)
		    }

		    if (state->blittingflags & DSBLIT_BLEND_COLORALPHA)		/* 0x00000002 - enables blending and uses alpha value from color */
		    {
			dev->cmd_opts &= ~GXA_CMD_NOT_ALPHA;
#warning check!
//			_write_gxa(drv->gxa_base, GXA_CFG2_REG, DEF_CFG2_REG | (state->color.a & 0xFF));
			_write_gxa(drv->gxa_base, GXA_CFG2_REG, DEF_CFG2_REG | (state->color.a));

			blend_reg = (0x0C << 16) | (0x0C << 12) | (0x0C << 6) | (0x08 << 2) | 0x02;	// (0x0C << 16) | (0x09 << 12) | (0x0C << 6)
//			blend_reg = (0x0C << 16) | (0x09 << 12) | (0x0C << 6);
		    }

		    if (state->blittingflags & DSBLIT_COLORIZE)			/* 0x00000004 - modulates source color with the color's r/g/b values */
		    {
			dev->cmd_opts &= ~GXA_CMD_NOT_ALPHA;

			if ((state->source->config.format == DSPF_A8) || (state->source->config.format == DSPF_LUT8))
			{
			    if (state->blittingflags & DSBLIT_BLEND_COLORALPHA)
			    {
				blend_reg = (0xFF << 24) |
					    (0x0C << 16) | // src blend ratio - color
					    (0x0D << 12) | // dst blend ratio - color
					    (0x0C <<  6) | // src blend ratio - alpha
					    (0x0D <<  2) | // dst blend ratio - alpha
					    (0x00);
			    }
			    else
			    {
				blend_reg = (0xFF << 24) |
					    (0x08 << 16) | // src blend ratio - color
					    (0x09 << 12) | // dst blend ratio - color
					    (0x00 <<  6) | // src blend ratio - alpha
					    (0x00 <<  2);  // dst blend ratio - alpha
			    }
			}
			else
			{
			    dev->b_soft = true;
//			    return;
			}
		    }

		    if (state->blittingflags & DSBLIT_SRC_COLORKEY)		/* 0x00000008 - don't blit pixels matching the source color key */
		    {
			dev->cmd_opts |= GXA_CMD_SRC_TRANSP;
                	cfg_reg |= 0x00005000;       /* Source Transparency enable & Source Transparency Alpha ignore */

		        /* setup the "chroma" keying registers */
			_write_gxa(drv->gxa_base, GXA_KEY_LOWER_REG, state->src_colorkey);
			_write_gxa(drv->gxa_base, GXA_KEY_UPPER_REG, state->src_colorkey);			
		    }

		    if (state->blittingflags & DSBLIT_DST_COLORKEY)		/* 0x00000010 - write to destination only if the destination pixel matches the destination color key */
		    {
			/* not supported yet */
		    }


		    if (state->blittingflags & DSBLIT_ROTATE180)
		    {
			/* not supported yet */
		    }

		    if (state->blittingflags)
		    {
			_write_gxa(drv->gxa_base, GXA_CFG_REG, cfg_reg);
			_write_gxa(drv->gxa_base, GXA_BLEND_CFG_REG, blend_reg);
		    }
		}

        	if (state->mod_hw & SMF_COLOR)		/* 0x00000008 */
	        {
		    /* for colorize DFB partly setups the hw only once and updates the color value itself */
//		    if (state->blittingflags & DSBLIT_COLORIZE)
//		    {
//        		switch (state->source->config.format)
//			{
//			    case DSPF_A8:
//			    {
				/* nothing to do here. A8 is handled in source vs dest config format check */
//				break;
//			    }
//			    case DSPF_ARGB:
//			    case DSPF_RGB32:
//		        	_write_gxa(drv->gxa_base, GXA_CONST_COLOR_REG, _make_cx2450x_color(state->color, state->destination->config.format));
//				break;
//			    default:
//				D_INFO("%s() unhandled config.format %s [0x%.8X]\n", __FUNCTION__, dfb_pixelformat_name(state->source->config.format), state->source->config.format);
//			}
//		    }

		    if (state->blittingflags & DSBLIT_SRC_COLORKEY)
		    {
		        _write_gxa(drv->gxa_base, GXA_KEY_LOWER_REG, state->src_colorkey);
			_write_gxa(drv->gxa_base, GXA_KEY_UPPER_REG, state->src_colorkey);
		    }

        	    _write_gxa(drv->gxa_base, GXA_CONST_COLOR_REG, _make_cx2450x_color(state->color, state->destination->config.format));
        	    _write_gxa(drv->gxa_base, GXA_BG_COLOR_REG, _make_cx2450x_color(state->color, state->destination->config.format));

		    if (state->blittingflags & ~CX2450X_SUPPORTED_BLITTING_FLAGS)
			D_INFO("%s() unhandled color update for blitting flag(s) %.8X\n", __FUNCTION__, state->blittingflags);
		}

		if (state->mod_hw & SMF_DESTINATION)	/* 0x00000100 */
		{
		    /* setup the destination bitmap - as supposed by the datasheet, we use bitmap context #2 */
	    	    _write_gxa(drv->gxa_base, GXA_BMP2_TYPE_REG, _make_cx2450x_bmp_reg(state->destination->config.format, state->dst.pitch));
		    _write_gxa(drv->gxa_base, GXA_BMP2_ADDR_REG, drv->smem_start + state->dst.offset);
		}
        	if (state->mod_hw & SMF_SOURCE)		/* 0x00000200 */
        	{
            	    /* use content bitmap #1 as source for blitting operations */
            	    _write_gxa(drv->gxa_base, GXA_BMP1_TYPE_REG, _make_cx2450x_bmp_reg(state->source->config.format, state->src.pitch));
            	    _write_gxa(drv->gxa_base, GXA_BMP1_ADDR_REG, drv->smem_start + state->src.offset);
        	}

    		if (state->source->config.format != state->destination->config.format)
    		{
        	    switch (state->source->config.format)
		    {
			case DSPF_RGB32:    /* nothing to do */
			case DSPF_ARGB:
			    break;
			case DSPF_RGB16:
			    _write_gxa(drv->gxa_base, GXA_CMD_PALETTE | GXA_SRC_BMP_SEL(1) | GXA_PARAM_COUNT(1), drv->RGB16_table);
			    break;
			case DSPF_A8:
			    if (state->blittingflags & DSBLIT_COLORIZE)
			    {
				/* modulate the grey-scale table with color */
				volatile u32 *pal_g = (volatile u32 *)drv->A8_base;
				volatile u32 *pal   = (volatile u32 *)drv->var_base;
				u32 i;

				for (i = 0; i < 256; i++)
				{
				    pal[i] = (((((pal_g[i] >> 24) & 0xFF) * ((int)state->color.a + 1)) & 0xFF00) << 16) |
					     (((((pal_g[i] >> 16) & 0xFF) * ((int)state->color.r + 1)) & 0xFF00) <<  8) |
					     (((((pal_g[i] >>  8) & 0xFF) * ((int)state->color.g + 1)) & 0xFF00)      ) |
				    	     (((((pal_g[i]      ) & 0xFF) * ((int)state->color.b + 1)) & 0xFF00) >>  8);
				}
				
				_write_gxa(drv->gxa_base, GXA_CMD_PALETTE | GXA_SRC_BMP_SEL(1) | GXA_PARAM_COUNT(1), drv->var_table);
			    }
			    else
				_write_gxa(drv->gxa_base, GXA_CMD_PALETTE | GXA_SRC_BMP_SEL(1) | GXA_PARAM_COUNT(1), drv->A8_table);
			    break;
			case DSPF_LUT8:
			{
            		    CorePalette *palette = state->source->palette;

            		    /* we have to exchange the palette values (DFB's view: ARGB, GXA's view: BGRA) */
            		    if (palette->num_entries)
			    {
				volatile u32 *pal = (volatile u32 *)drv->var_base;
				u32 i;

				if (state->blittingflags & DSBLIT_COLORIZE)
				{
				    D_INFO("Palette for COLORIZE\n");
				    for (i = 0; i < palette->num_entries; i++)
					pal[i] = PIXEL_ARGB( palette->entries[i].a, 
							    (palette->entries[i].r * (state->color.r + 1)) >> 8, 
							    (palette->entries[i].g * (state->color.g + 1)) >> 8, 
							    (palette->entries[i].b * (state->color.b + 1)) >> 8);
				}
				else
				{
				    D_INFO("Palette for NON colorize\n");
				    for (i = 0; i < palette->num_entries; i++)
					pal[i] = PIXEL_ARGB(palette->entries[i].a, palette->entries[i].r, palette->entries[i].g, palette->entries[i].b);
				}

				_write_gxa(drv->gxa_base, GXA_CMD_PALETTE | GXA_SRC_BMP_SEL(1) | GXA_PARAM_COUNT(1), drv->var_table);
			    }
			    else
				D_INFO("%s() given LUT8 has no entries!\n", __FUNCTION__);
			    break;
			}
	
			default:
			    D_INFO("%s() unhandled color space conversion (table missing)\n", __FUNCTION__);
		    }	/* switch (state->source->config.format) */
		    break;
		}
		break;
	    }	/* case DFXL_BLIT */

	    default:
	        D_INFO("%s() unhandled acceleration mask 0x%.8X\n", __FUNCTION__, accel);

	}	/* switch(accel) */

	state->mod_hw = 0;

    }	/* if (state->mod_hw) */

}
