/*
 * Conexant/NXP/Trident CX2450x accelerated driver
 *
 * (c) Coolstream International Limited
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>

#include <directfb.h>

#include <direct/mem.h>
#include <direct/messages.h>

#include <core/gfxcard.h>
#include <core/graphics_driver.h>
#include <core/state.h>
#include <core/surface.h>
#include <core/palette.h>
#include <gfx/convert.h>

#include <fbdev/fbdev.h>

#include "cx2450x.h"
#include "cnxtfb.h"
#include "cx2450x-tables.h"

DFB_GRAPHICS_DRIVER(cx2450x);

#define CX2450X_SUPPORTED_DRAWING_FUNCTIONS	(DFXL_FILLRECTANGLE | DFXL_DRAWRECTANGLE | DFXL_DRAWLINE)
#define CX2450X_SUPPORTED_DRAWING_FLAGS		(DSDRAW_BLEND)

#define CX2450X_SUPPORTED_BLITTING_FUNCTIONS	(DFXL_BLIT | DFXL_STRETCHBLIT)
//#define CX2450X_SUPPORTED_BLITTING_FLAGS	(DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_BLEND_COLORALPHA | DSBLIT_COLORIZE | DSBLIT_SRC_COLORKEY)
#define CX2450X_SUPPORTED_BLITTING_FLAGS	(DSBLIT_BLEND_COLORALPHA | DSBLIT_COLORIZE | DSBLIT_SRC_COLORKEY)

static u32 gxa_cfg_reg;
static u32 gxa_cfg2_reg;
static u32 gxa_line_control_reg;
static u32 gxa_bg_color_reg;
static u32 gxa_fg_color_reg;
static u32 gxa_const_color_reg;
static u32 gxa_blend_cfg_reg;
static u32 gxa_key_lower_reg;
static u32 gxa_key_upper_reg;

static CorePalette *private_pal;

static u32 _make_cx2450x_bmp_reg(DFBSurfacePixelFormat dfbcolorformat, u32 pitch);

static void _cx2450x_check_state(void *driver_data, void *device_data, CardState *state, DFBAccelerationMask accel);
static void _cx2450x_set_state(void *driver_data, void *device_data, GraphicsDeviceFuncs *funcs, CardState *state, DFBAccelerationMask accel);

bool _cx2450x_draw_line(void *driver_data, void *device_data, DFBRegion *line);
static bool _cx2450x_draw_rectangle(void *driver_data, void *device_data, DFBRectangle *rect);
static bool _cx2450x_fill_rectangle(void *driver_data, void *device_data, DFBRectangle *rect);
//static bool _cx2450x_fill_triangle(void *driver_data, void *device_data, DFBTriangle *tri);
static bool _cx2450x_blit(void *driver_data, void *device_data, DFBRectangle *rect, int dx, int dy);
static bool _cx2450x_stretchblit( void *driver_data, void *device_data, DFBRectangle *srect, DFBRectangle *drect);

#include "cx2450x-accel.c"	/* contains the acceleration functions */
#include "cx2450x-state.c"	/* contains set/check state functions */
#include "cx2450x-osd.c"	/* contains OSD (DisplayLAyer) functions */

/*******************************************************************************/

static int driver_probe(CoreGraphicsDevice *device)
{
    int x;

    x = dfb_gfxcard_get_accelerator(device);

    if (x == FB_ACCEL_CX2450X)
	return 1;
    return 0;
}

/*******************************************************************************/

static void driver_get_info(CoreGraphicsDevice *device, GraphicsDriverInfo *info)
{
    (void)device;

    snprintf(info->name,   DFB_GRAPHICS_DRIVER_INFO_NAME_LENGTH,   "CX2450X GFX Driver");
    snprintf(info->vendor, DFB_GRAPHICS_DRIVER_INFO_VENDOR_LENGTH, "Coolstream International Limited");

    info->version.major = 0;
    info->version.minor = 2;

    info->driver_data_size = sizeof(CX2450XDriverData);
    info->device_data_size = sizeof(CX2450XDeviceData);
}

/*******************************************************************************/

static DFBResult driver_init_driver(CoreGraphicsDevice *device, GraphicsDeviceFuncs *funcs, void *driver_data, void *device_data, CoreDFB *core)
{

    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
//    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;
    volatile u32      *reg;
    u32		       val;

    FBDev *dfb_fbdev;

    drv->gxa_base = (void*) -1;

    dfb_fbdev = dfb_system_data();
    if (!dfb_fbdev)
    	return DFB_IO;

    drv->smem_start  = dfb_fbdev->shared->fix.smem_start;
    drv->smem_len    = dfb_fbdev->shared->fix.smem_len;/* - 0x00040000; */	/* reserve 256 KB for private data (conversion tables) */
    drv->priv_start  = drv->smem_start + drv->smem_len;
    drv->priv_len    = 0x040000;
    drv->coeff_table = drv->priv_start;
    drv->A8_table    = drv->priv_start + 0x08000;
    drv->RGB16_table = drv->priv_start + 0x20000;
    drv->var_table   = drv->priv_start + 0x30000;

    drv->devmem_fd = open("/dev/mem", O_RDWR | O_SYNC);
    if (drv->devmem_fd < 0)
	return DFB_IO;

    drv->gxa_base = mmap(NULL, 0x00040000, PROT_READ | PROT_WRITE, MAP_SHARED, drv->devmem_fd, 0xE0600000);
    if (drv->gxa_base == (void*) -1)
	return DFB_IO;

    drv->drm_base = mmap(NULL, 0x00001000, PROT_READ | PROT_WRITE, MAP_SHARED, drv->devmem_fd, 0xE0560000);
    if (drv->drm_base == (void*) -1)
	return DFB_IO;

    reg = (volatile u32*)(drv->drm_base + 0x00000080);
    val = *reg;
    drv->osd_base = mmap(NULL, 0x00010000, PROT_READ | PROT_WRITE, MAP_SHARED, drv->devmem_fd, val & 0xFFFF0000);
    if (drv->osd_base == (void*) -1)
	return DFB_IO;
    drv->osd_ofs = val & 0x0000FFFF;

    reg = (volatile u32*)(drv->osd_base + drv->osd_ofs + 0x0000001C);
    val = *reg;

    drv->pal_base = mmap(NULL, 0x00010000, PROT_READ | PROT_WRITE, MAP_SHARED, drv->devmem_fd, val & 0xFFFF0000);
    if (drv->pal_base == (void*) -1)
	return DFB_IO;
    drv->pal_ofs = val & 0x0000FFFF;

    drv->coeff_base = mmap(NULL, 0x00010000, PROT_READ | PROT_WRITE, MAP_SHARED, dfb_fbdev->fd, drv->smem_len);
    if (drv->coeff_base == (void*) -1)
	return DFB_IO;

    drv->A8_base = mmap(NULL, 0x00008000, PROT_READ | PROT_WRITE, MAP_SHARED, dfb_fbdev->fd, drv->smem_len + 0x08000);
    if (drv->A8_base == (void*) -1)
	return DFB_IO;

    drv->RGB16_base = mmap(NULL, 0x00010000, PROT_READ | PROT_WRITE, MAP_SHARED, dfb_fbdev->fd, drv->smem_len + 0x20000);
    if (drv->RGB16_base == (void*) -1)
	return DFB_IO;

    drv->var_base = mmap(NULL, 0x00010000, PROT_READ | PROT_WRITE, MAP_SHARED, dfb_fbdev->fd, drv->smem_len + 0x30000);
    if (drv->var_base == (void*) -1)
	return DFB_IO;

    /* copy the coefficients and luts */
    memcpy((u8*)drv->coeff_base, scale_coeff_table       , 0x0400);
    memcpy((u8*)drv->A8_base   , DSPF_A8_to_ARGB_table   , 1024);
    memcpy((u8*)drv->RGB16_base, DSPF_RGB16_to_ARGB_table, 256);

    /* base functions */
    funcs->EngineSync = _cx2450x_sync_engine;
    funcs->CheckState = _cx2450x_check_state;
    funcs->SetState   = _cx2450x_set_state;

    /* drawing functions */
    funcs->DrawLine      = _cx2450x_draw_line;
    funcs->DrawRectangle = _cx2450x_draw_rectangle;
    funcs->FillRectangle = _cx2450x_fill_rectangle;
    funcs->FillTriangle  = NULL; //_cx2450x_fill_triangle;

    /* blitting functions */
    funcs->Blit        = _cx2450x_blit;
    funcs->StretchBlit = _cx2450x_stretchblit;

    /* set some blender config (OSD plane at a2 and a3) */
    reg   = (volatile u32*)(drv->drm_base + 0x00000054);
    *reg &= 0xFFFFFFFD;

    reg   = (volatile u32*)(drv->osd_base + drv->osd_ofs + 0x00000008);
    *reg &= 0x1FFFFFFF;
    *reg |= 0x40000000;

    /* we need a palette for our hook. Generate it here globally */
    dfb_palette_create(core, 256, &private_pal);
    
    /* finally install primary layer hooks */
    dfb_layers_hook_primary(device, driver_data, &newPrimaryFuncs, &oldPrimaryFuncs, &oldPrimaryDriverData);

    return DFB_OK;
}

/*******************************************************************************/

static DFBResult driver_init_device(CoreGraphicsDevice *device, GraphicsDeviceInfo *device_info, void *driver_data, void *device_data)
{
//    u32 reg;
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;
//    CX2450XDeviceData *dev = (CX2450XDeviceData *)device_data;

    snprintf(device_info->name,   DFB_GRAPHICS_DEVICE_INFO_NAME_LENGTH,   "CX2450x");
    snprintf(device_info->vendor, DFB_GRAPHICS_DEVICE_INFO_VENDOR_LENGTH, "NXP");

    device_info->caps.flags    = 0;										/* CardCapabilitiesFlags */
    device_info->caps.accel    = CX2450X_SUPPORTED_DRAWING_FUNCTIONS | CX2450X_SUPPORTED_BLITTING_FUNCTIONS;	/* DFBAccelerationMask */
    device_info->caps.blitting = CX2450X_SUPPORTED_BLITTING_FLAGS;						/* DFBSurfaceBlittingFlags */
    device_info->caps.drawing  = CX2450X_SUPPORTED_DRAWING_FLAGS;						/* DFBSurfaceDrawingFlags */
    device_info->caps.clip     = DFXL_NONE;									/* DFBAccelerationMask */

    device_info->limits.surface_byteoffset_alignment        = 128;	// ??? /* unsigned int */
    device_info->limits.surface_pixelpitch_alignment        = 32;	// ??? /* unsigned int */

    device_info->limits.surface_bytepitch_alignment         = 4;	/* unsigned int */

    /* save some registers to be able to restore on exit */
    gxa_cfg_reg		 = _read_gxa(drv->gxa_base, GXA_CFG_REG);
    gxa_cfg2_reg	 = _read_gxa(drv->gxa_base, GXA_CFG2_REG);
    gxa_line_control_reg = _read_gxa(drv->gxa_base, GXA_LINE_CONTROL_REG);
    gxa_bg_color_reg	 = _read_gxa(drv->gxa_base, GXA_BG_COLOR_REG);
    gxa_fg_color_reg	 = _read_gxa(drv->gxa_base, GXA_FG_COLOR_REG);
    gxa_const_color_reg	 = _read_gxa(drv->gxa_base, GXA_CONST_COLOR_REG);
    gxa_blend_cfg_reg	 = _read_gxa(drv->gxa_base, GXA_BLEND_CFG_REG);
    gxa_key_lower_reg	 = _read_gxa(drv->gxa_base, GXA_KEY_LOWER_REG);
    gxa_key_upper_reg	 = _read_gxa(drv->gxa_base, GXA_KEY_UPPER_REG);

    /* setup config registers with our own defaults */
    _write_gxa(drv->gxa_base, GXA_CFG_REG, DEF_CFG_REG);
    _write_gxa(drv->gxa_base, GXA_CFG2_REG, DEF_CFG2_REG);

    /* disable all content lock */
    _write_gxa(drv->gxa_base, GXA_CONTENT_ID_REG, 0);

    return DFB_OK;
}

/*******************************************************************************/

static void driver_close_device(CoreGraphicsDevice *device, void *driver_data,	void *device_data)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;

    /* restore some registers */
    _write_gxa(drv->gxa_base, GXA_CFG_REG ,gxa_cfg_reg);
    _write_gxa(drv->gxa_base, GXA_CFG2_REG ,gxa_cfg2_reg);
    _write_gxa(drv->gxa_base, GXA_LINE_CONTROL_REG ,gxa_line_control_reg);
    _write_gxa(drv->gxa_base, GXA_BG_COLOR_REG ,gxa_bg_color_reg);
    _write_gxa(drv->gxa_base, GXA_FG_COLOR_REG ,gxa_fg_color_reg);
    _write_gxa(drv->gxa_base, GXA_CONST_COLOR_REG ,gxa_const_color_reg);
    _write_gxa(drv->gxa_base, GXA_BLEND_CFG_REG ,gxa_blend_cfg_reg);
    _write_gxa(drv->gxa_base, GXA_KEY_LOWER_REG ,gxa_key_lower_reg);
    _write_gxa(drv->gxa_base, GXA_KEY_UPPER_REG ,gxa_key_upper_reg);
}

/*******************************************************************************/

static void driver_close_driver(CoreGraphicsDevice *device, void *driver_data)
{
    CX2450XDriverData *drv = (CX2450XDriverData *)driver_data;

    if (drv->gxa_base != (void*) -1)
	munmap((void*) drv->gxa_base, 0x00040000);

    if (drv->drm_base != (void*) -1)
	munmap((void*) drv->drm_base, 0x00001000);

    if (drv->osd_base != (void*) -1)
	munmap((void*) drv->osd_base, 0x00010000);

    if (drv->pal_base != (void*) -1)
	munmap((void*) drv->pal_base, 0x00010000);

    if (drv->coeff_base != (void*) -1)
	munmap((void*) drv->coeff_base, 0x00010000);

    if (drv->A8_base != (void*) -1)
	munmap((void*) drv->A8_base, 0x00010000);

    if (drv->RGB16_base != (void*) -1)
	munmap((void*) drv->RGB16_base, 0x00010000);

    if (drv->var_base != (void*) -1)
	munmap((void*) drv->var_base, 0x00010000);

    if (drv->devmem_fd > -1)
	close(drv->devmem_fd);
}
