/*
 * Conexant / NXP CX2450x accelated driver
 *
 * (c) Coolstream International Limited
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

#ifndef __CX2450X_H__
#define __CX2450X_H__

#include <direct/types.h>

D_DEBUG_DOMAIN( cx2450x, "CX2450x", "Trident CX2450x gfx driver" );

/* GXA register definitions (as offset) */
#define GXA_FG_COLOR_REG	0x0020
#define GXA_BG_COLOR_REG	0x0024
#define GXA_LINE_PATT_REG	0x0028
#define	GXA_CFG_REG		0x0030
#define GXA_BLT_CONTROL_REG	0x0034
#define GXA_LINE_CONTROL_REG	0x0038
#define GXA_BLEND_CFG_REG	0x003C

#define	GXA_BMP0_TYPE_REG	0x0040
#define	GXA_BMP0_ADDR_REG	0x0044

#define	GXA_BMP1_TYPE_REG	0x0048
#define	GXA_BMP1_ADDR_REG	0x004C

#define	GXA_BMP2_TYPE_REG	0x0050
#define	GXA_BMP2_ADDR_REG	0x0054

#define	GXA_BMP3_TYPE_REG	0x0058
#define	GXA_BMP3_ADDR_REG	0x005C

#define	GXA_BMP4_TYPE_REG	0x0060
#define	GXA_BMP4_ADDR_REG	0x0064

#define	GXA_BMP5_TYPE_REG	0x0068
#define	GXA_BMP5_ADDR_REG	0x006C

#define	GXA_BMP6_TYPE_REG	0x0070
#define	GXA_BMP7_TYPE_REG	0x0078

#define GXA_DEPTH_REG		0x00F4
#define	GXA_CFG2_REG		0x00FC
#define GXA_SC_NORM_REG		0x0120
#define GXA_CSC_SET0_REG	0x0124
#define GXA_CSC_SET1_REG	0x0128
#define GXA_CSC_SET2_REG	0x012C
#define	GXA_CONST_COLOR_REG	0x0134
#define	GXA_KEY_UPPER_REG	0x0138
#define	GXA_KEY_LOWER_REG	0x013C
#define GXA_CONTENT_ID_REG	0x0144

/* GXA commands at bit 17:11 */
#define GXA_CMD_QMARK		0x00001000
#define	GXA_CMD_PALETTE		0x00001800
#define GXA_CMD_VFILTER		0x00002000
#define GXA_CMD_HFILTER		0x00002800
#define GXA_CMD_BLT		0x00010800	/* perform blitting if set, or line draw if not set */
#define GXA_CMD_NOT_ALPHA	0x00011000	/* no alpha blending if set */
#define GXA_CMD_SRC_TRANSP	0x00012000	/* use source transparency if set */
#define GXA_CMD_ROP		0x00014000	/* special Raster Operations if set (type of ROP must be setup manually), or simply copy if not set */
#define GXA_CMD_NOT_TEXT	0x00018000	/* no special text blit algorythm if set (only valid for patterns) */
#define GXA_CMD_STRETCH_BLT	0x00030000	/* stretch blit if enabled */

/* various bitfield definitions for GXA addressing (if it's not a register R/W) */
#define GXA_SRC_BMP_SEL(x)	(x << 8)	/* Source bitmap selector from 0 to 7 (for blitting cmds) */
#define GXA_DST_BMP_SEL(x)	(x << 5)	/* destination bitmap selector from 0 to 7 (for blitting and drawing cmds) */
#define GXA_PARAM_COUNT(x)	(x << 2)	/* command parameter from 0 to 7 */

/* encode pixel data into register values */
#define GXA_POINT(x, y)		(((y) << 16) | (x))

/*defaults for config registers */
#define DEF_CFG_REG		0x00840000
#define DEF_CFG2_REG		0x00000200

/* defaults for GXA BLEND */
#define DEF_BLEND_REG_PIX   ((0x01 << 24) | (0x08 << 16) | (0x09 << 12) | (0x08 << 6))	/* blending with source pixel alpha */
#define DEF_BLEND_REG_FIX   ((0x01 << 24) | (0x0C << 16) | (0x09 << 12) | (0x0C << 6))	/* blending with fixed alpha in CFG2 */

/* OSD Header definitions (as offset) */
#define OSD_NEXT_PONTER		0x00
#define OSD_X_DISP_START_WIDTH	0x04
#define OSD_IMG_HEIGHT_WIDTH	0x08
#define OSD_GDMA		0x0C
#define OSD_LINE_STRIDE_BITF	0x10
#define OSD_Y_POS_REGION_ALPHA	0x14
#define OSD_SCALE_FACTORS	0x18
#define OSD_PALETTE_ADDRESS	0x1C
#define OSD_CHROMA_KEY_MIN	0x20
#define OSD_CHROMA_KEY_MAX	0x24

/* OSD Header bit definitions */
/* for OSD_IMG_HEIGHT_WIDTH */
#define OSD_ALPHA_MULT_EN	0x10000000

/*for OSD_LINE_STRIDE_BITF */
#define OSD_ALPHA_BL_BOTH	0x40000000
#define OSD_ALPHA_BL_FORCE	0x20000000
#define OSD_FLICKER_FILTER	0x01000000
#define OSD_PALETTE_LOAD	0x00800000


typedef struct
{
    unsigned long	 smem_start;		/* physical start address of the framebuffer */
    unsigned int	 smem_len;		/* length of the framebuffer */
    int			 devmem_fd;		/* the filedescriptor for the opened /dev/mem */
    volatile u8		*gxa_base;		/* the mmap'd GXA register address space */
    volatile u8		*drm_base;		/* the mmap'd DRM register address space to get osd config */
    volatile u8		*osd_base;		/* the mmap'd DRM register address space for blend config */
    volatile u8		*pal_base;		/* even for 32 bit RGB the HW uses a palette (for alpha) */
    volatile u8		*coeff_base;		/* the mmap'd view of the private used space */
    volatile u8		*A8_base;
    volatile u8		*RGB16_base;
    volatile u8		*var_base;
    u32	 		 priv_start;		/* the start address of private used framebuffer space (not seen by DFB) */
    u32			 priv_len;		/* the size of private used space */
    u32			 coeff_table;
    u32			 A8_table;
    u32			 RGB16_table;
    u32			 var_table;
    u32			 osd_ofs;		/* mmap works on boundaries, so this is the additional offset to the OSD header */
    u32			 pal_ofs;		/* mmap works on boundaries, so this is the additional offset to the palette */
} CX2450XDriverData;

typedef struct
{
    u32		 		cmd_opts;	/* additional command options */
    DFBSurfaceBlittingFlags	b_flags;	/* local copy of blitting flags from state->mod_hw to make workarounds for DFB-bugs */
    bool			b_soft;
} CX2450XDeviceData;

#endif /* __CX2450X_H__ */
